"""API Importer Protocol and utilities for custom importers"""

import abc
from datetime import date
from typing import Any

from beancount.core.data import Amount, Transaction

TransactionReport = tuple[list[Transaction], Amount]


def prepare_meta(d: dict[str, Any]) -> dict[str, str]:
    """Return a dict of metadata for a Beancount transaction.

    Args:
        d (dict[str, Any]): a dict of values

    Returns:
        dict[str, str]: a dict of metadata
    """
    meta = {}
    for k, v in d.items():
        if not (v is None or v == ""):
            meta[k] = str(v)
    return meta


class ApiImporterProtocol(abc.ABC):
    """API Importer Protocol for custom importers

    All API importers must comply with this interface. Make sure to implement
    all methods decorated with `@abc.abstractmethod`. There are no restrictions
    on other methods, variables or properties.
    """

    @abc.abstractmethod
    def fetch_transactions(
        self,
        bean_account: str,
        from_date: date,
        to_date: date,
    ) -> TransactionReport:
        """Return a tuple with the list of transactions and the current balance.

        Args:
            bean_account (str): a Beancount account name
            from_date (date): the first date to import
            to_date (date): the last date to import

        Returns:
            TransactionReport: A tuple with the list of transactions and
                the current balance.
        """



"""Beanclerk command-line interface"""

from datetime import date
from pathlib import Path

import click

from .clerk import import_transactions
from .exceptions import BeanclerkError

CONFIG_FILE = "beanclerk-config.yml"


@click.group(
    context_settings={"help_option_names": ["-h", "--help"]},
    no_args_is_help=False,
)
@click.option(
    "-c",
    "--config-file",
    default=Path.cwd() / CONFIG_FILE,
    type=click.Path(path_type=Path),
    help=f"Path to a config file; defaults to `{CONFIG_FILE}` in the current working directory.",  # noqa: E501
)
@click.pass_context
def cli(ctx: click.Context, config_file: Path) -> None:
    """Automation for Beancount.

    Import and categorize transactions via API importers and user-defined rules.
    """
    # https://click.palletsprojects.com/en/8.1.x/commands/#nested-handling-and-contexts
    ctx.ensure_object(dict)
    ctx.obj["config_file"] = config_file


_ISO_DATE_FMT: str = "YYYY-MM-DD"


class Date(click.ParamType):
    """A convenience date type for Click.

    Converts dates to a date (instead of datetime).
    """

    name = "date"

    def convert(self, value, param, ctx):
        if isinstance(value, date):
            return value
        try:
            return date.fromisoformat(value)
        except ValueError:
            self.fail(
                f"'{value}' is not a valid date format ({_ISO_DATE_FMT})",
                param,
                ctx,
            )


@cli.command("import")
@click.option(
    "--from-date",
    type=Date(),
    help=f"The first date to import ({_ISO_DATE_FMT}).",
)
@click.option(
    "--to-date",
    type=Date(),
    help=f"The last date to import ({_ISO_DATE_FMT}).",
)
@click.pass_context
def import_(ctx: click.Context, from_date: Date, to_date: Date) -> None:
    """Import transactions and check the current balance."""
    try:
        import_transactions(
            config_file=ctx.obj["config_file"],
            from_date=from_date,
            to_date=to_date,
        )
    except BeanclerkError as exc:
        raise click.ClickException(str(exc)) from exc



"""Custom exceptions"""


class BeanclerkError(Exception):
    """Base class for all Beanclerk exceptions"""


class ConfigError(BeanclerkError):
    """Error in the Beanclerk configuration file"""

    def __init__(self, message: str) -> None:
        super().__init__(f"Cannot load config file: {message}")


class ClerkError(BeanclerkError):
    """An Error during clerk operations"""

    def __init__(self, message: str) -> None:
        super().__init__(f"Clerk error: {message}")



"""Banka Creditas a.s.

TODO:
    This module is a work in progress. It needs a major rework.

docs:
    https://www.creditas.cz/firma/creditas-api/
"""

import base64
from datetime import date
from decimal import Decimal

import creditas
from beancount.core.data import Amount, Transaction
from beancount.core.flags import FLAG_WARNING
from lxml import etree

from ..bean_helpers import create_posting, create_transaction
from . import prepare_meta


def _get_transactions(token: str, account_id: str, from_date: date) -> bytes:
    # Creditas API v 1.0.0:
    # Manually generated token can be used only with the following URLs (and
    # the corresponding methods):
    #   /account/current/get
    #   /account/savings/get
    #   /account/balance/get
    #   /account/transaction/search
    #   /account/transaction/export
    #   /account/statement/list
    #   /account/statement/get

    config = creditas.Configuration()
    config.access_token = token
    api = creditas.TransactionApi(creditas.ApiClient(config))
    body = creditas.Body8(
        account_id=account_id,
        format="XML",
        filter=creditas.AccountTransactionFilter(date_from=from_date),
    )
    # TODO: handle creditas.rest.ApiException
    data: creditas.InlineResponse20011 = api.d_ps_account_transaction_export_api(
        body=body,
    )
    # TODO: handle other exceptions
    return base64.b64decode(data.export)


# FIXME: This function has to be turned into a class implementing ApiImporterProtocol.
def get_transactions(
    token: str,
    account_id: str,
    bean_account: str,
    from_date: date,
) -> tuple[list[Transaction], Amount]:
    # TODO: handle etree.XMLSyntaxError
    # TODO: handle other exceptions
    xml_root = etree.fromstring(  # noqa: S320
        _get_transactions(token, account_id, from_date),
    )
    nsmap = xml_root.nsmap

    # TODO: handle other exceptions
    # TODO: raise if not found
    def get_amount(element) -> Amount:
        amount = element.find("./Amt", nsmap)
        number = Decimal(amount.text)
        currency = amount.attrib["Ccy"]
        if element.find("./CdtDbtInd", nsmap).text == "DBIT":
            number = -number
        return Amount(number, currency)

    # TODO: handle other exceptions
    def get_text(element, xpath: str, *, raise_if_none: bool = False) -> str | None:
        text: str | None = element.findtext(xpath, default=None, namespaces=nsmap)
        if raise_if_none and text is None:
            # TODO: raise a custom exception
            raise
        return text

    statement = xml_root.find("./BkToCstmrStmt/Stmt", nsmap)
    balance = get_amount(statement.find("./Bal", nsmap))
    num_entries = get_text(
        statement,
        "./TxsSummry/TtlNtries/NbOfNtries",
        raise_if_none=True,
    )
    if num_entries == 0:
        return ([], balance)
    txns: list[Transaction] = []
    for entry in statement.findall("./Ntry", nsmap):
        if get_text(entry, "./CdtDbtInd", raise_if_none=True) == "DBIT":
            # Whether the related party is a debitor or a creditor.
            ind = "Cdtr"
        else:
            ind = "Dbtr"
        details = "./NtryDtls/TxDtls"
        # TODO: change `transaction_id` to `id` and `recipient_message`
        #   to `remittance_info`
        meta = prepare_meta(
            {
                "transaction_id": get_text(entry, "./NtryRef", raise_if_none=True),
                "account_id": get_text(
                    entry,
                    f"{details}/RltdPties/{ind}Acct/Id/Othr/Id",
                ),
                "bank_id": get_text(
                    entry,
                    f"{details}/RltdAgts/{ind}Agt/FinInstnId/Othr/Id",
                ),
                "ks": get_text(entry, f"{details}/Refs/InstrId"),
                "vs": get_text(entry, f"{details}/Refs/EndToEndId"),
                "ss": get_text(entry, f"{details}/Refs/PmtInfId"),
                "recipient_message": get_text(entry, f"{details}/RmtInf/Ustrd"),
                "executor": get_text(entry, f"{details}/RltdPties/{ind}/Nm"),
            },
        )
        txns.append(
            create_transaction(
                _date=date.fromisoformat(
                    get_text(
                        entry,
                        "./BookgDt/Dt",
                        raise_if_none=True,
                    ),  # type: ignore[arg-type]
                ),
                flag=FLAG_WARNING,
                postings=[
                    create_posting(
                        account=bean_account,
                        units=get_amount(entry),
                    ),
                ],
                meta=meta,
            ),
        )
    return (txns, balance)



"""Tests of the config module."""
from beanclerk.config import load_config, load_importer
from beanclerk.importers.fio_banka import ApiImporter


def test_load_config(config_file, ledger):
    """Test load_config."""
    load_config(config_file)  # raises on invalid config


def test_load_importer(config_file, ledger):
    """Test load_importer."""
    config = load_config(config_file)
    for account_config in config.accounts:
        importer = load_importer(account_config)  # raises on invalid config
        assert isinstance(importer, ApiImporter)



"""Helpers for Beancount"""
from collections.abc import Generator
from datetime import date
from typing import TypeVar

from beancount.core.account import is_valid
from beancount.core.data import (
    EMPTY_SET,
    Account,
    Amount,
    Cost,
    CostSpec,
    Directive,
    Flag,
    Meta,
    Posting,
    Transaction,
)
from beancount.core.flags import FLAG_OKAY


def create_transaction(
    _date: date,
    flag: Flag = FLAG_OKAY,
    payee: str | None = None,
    narration: str = "",
    tags: frozenset | None = None,
    links: frozenset | None = None,
    postings: list[Posting] | None = None,
    meta: Meta | None = None,
) -> Transaction:
    """Return Transaction."""
    return Transaction(
        meta=meta if meta is not None else {},
        date=_date,
        flag=flag,
        payee=payee,
        narration=narration,
        tags=tags if tags is not None else EMPTY_SET,
        links=links if links is not None else EMPTY_SET,
        postings=postings if postings is not None else [],
    )


def create_posting(
    account: Account,
    units: Amount,
    cost: Cost | CostSpec | None = None,
    price: Amount | None = None,
    flag: Flag | None = None,
    meta: Meta | None = None,
) -> Posting:
    """Return Posting."""
    return Posting(
        account=account,
        units=units,
        cost=cost,
        price=price,
        flag=flag,
        meta=meta if meta is not None else {},
    )


D = TypeVar("D", bound=Directive)


def filter_entries(entries: list[Directive], cls: D) -> Generator[D, None, None]:
    """Yield only instances of a given Beancount directive.

    Args:
        entries (list[Directive]): a list of Beancount directives
        cls (Directive): a Beancount directive class

    Yields:
        Directive: a Beancount directive
    """
    for entry in entries:
        if isinstance(entry, cls):
            yield entry


def check_account_name(name: str) -> None:
    """Check a Beanount account name.

    Args:
        name (str): a Beancount account name

    Raises:
        ValueError: if the account name is invalid
    """
    if not is_valid(name):
        raise ValueError(f"'{name}' is not a valid Beancount account name")



"""API importer for Fio banka, a.s.

TODO:
    * handle fio_banka exceptions

docs:
    https://www.fio.cz/bank-services/internetbanking-api
"""

from datetime import date

import fio_banka
from beancount.core.data import Amount, Transaction

from ..bean_helpers import create_posting, create_transaction
from . import ApiImporterProtocol, TransactionReport, prepare_meta


class ApiImporter(ApiImporterProtocol):
    """API importer for Fio banka, a.s.

    Args:
        token (str): Fio banka API token
    """

    def __init__(self, token: str) -> None:
        self._token = token

    def fetch_transactions(
        self,
        bean_account: str,
        from_date: date,
        to_date: date,
    ) -> TransactionReport:
        client = fio_banka.Account(self._token)
        statement = client.periods(from_date, to_date, fio_banka.TransactionsFmt.JSON)

        txns: list[Transaction] = []
        for txn in fio_banka.get_transactions(statement):
            txns.append(
                create_transaction(
                    _date=txn.date,
                    postings=[
                        create_posting(
                            account=bean_account,
                            units=Amount(txn.amount, txn.currency),
                        ),
                    ],
                    meta=prepare_meta(
                        {
                            "id": txn.transaction_id,
                            "account_id": txn.account,
                            "account_name": txn.account_name,
                            "bank_id": txn.bank_id,
                            "bank_name": txn.bank_name,
                            "ks": txn.ks,
                            "vs": txn.vs,
                            "ss": txn.ss,
                            "user_identification": txn.user_identification,
                            "remittance_info": txn.recipient_message,
                            "type": txn.type,
                            "executor": txn.executor,
                            "specification": txn.specification,
                            "comment": txn.comment,
                            "bic": txn.bic,
                            "order_id": txn.order_id,
                            "payer_reference": txn.payer_reference,
                        },
                    ),
                ),
            )

        account_info = fio_banka.get_account_info(statement)
        return (txns, Amount(account_info.closing_balance, account_info.currency))



"""Configuration file parsing and validation.

Notes:
    * If a validator modifies a value, it should always return the same type:
    https://github.com/pydantic/pydantic/discussions/3997

TODO:
    * Add missing field validators.
"""
# Disabling due to Pydantic notation (`cls` instead of `self`).
# ruff: noqa: N805

import importlib
import os
from pathlib import Path
from typing import Any

import yaml
from pydantic import BaseModel, ConfigDict, ValidationError, field_validator
from pydantic_settings import BaseSettings

from .bean_helpers import check_account_name
from .exceptions import ConfigError
from .importers import ApiImporterProtocol


class BaseModelStrict(BaseModel):
    model_config = ConfigDict(extra="forbid")


class AccountConfig(BaseModel):
    """Account config model

    Allows extra fields to support custom configuration of importers.
    """

    account: str
    importer: str

    model_config = ConfigDict(extra="allow")  # allows access to extra fields

    @field_validator("account")
    def name_is_valid(cls, name: str) -> str:
        check_account_name(name)
        return name


class MatchCategories(BaseModelStrict):
    metadata: dict[str, str]


class CategorizationRule(BaseModelStrict):
    matches: MatchCategories
    account: str
    flag: str | None = None
    payee: str | None = None
    narration: str | None = None


class Config(BaseSettings):
    """Beanclerk config

    Most attributes are defined in a config file. Config is a Pydantic
    model, and raises a `pydantic.ValidationError` on invalid fields.
    """

    vars: Any = None  # noqa: A003
    input_file: Path
    accounts: list[AccountConfig]
    categorization_rules: list[CategorizationRule] | None = None

    # fields not present in the config file
    config_file: Path

    model_config = ConfigDict(extra="forbid")

    @field_validator("input_file")
    def input_file_exists(cls, input_file: Path) -> Path:
        # Side effects:
        #   * expands user (`~`) and environment variables
        filename: str = os.path.expandvars(input_file.expanduser())
        if not os.path.isabs(filename):  # noqa: PTH117
            filename = os.path.normpath(Path.cwd() / filename)
        input_file = Path(filename)
        if not input_file.exists():
            raise ValueError(f"Input file '{input_file}' does not exist")
        return input_file


def load_config(filepath: Path) -> Config:
    """Load and validate and a Beanclerk config file object.

    Args:
        filepath (Path): path to a config file

    Raises:
        ConfigError: Raised when the config file cannot be loaded or is invalid

    Returns:
        Config: a validated config object
    """
    try:
        with filepath.open("r") as file:
            contents = yaml.safe_load(file)
            contents["config_file"] = filepath
            return Config.model_validate(contents)
    except (OSError, yaml.YAMLError, ValidationError) as exc:
        raise ConfigError(str(exc)) from exc


def load_importer(account_config: AccountConfig) -> ApiImporterProtocol:
    """Return an instance of importer defined in the account config.

    Args:
        account_config (AccountConfig)

    Raises:
        ConfigError: Raised when the importer cannot be loaded

    Returns:
        ApiImporterProtocol: an instance of a particular importer implementing
            the API Importer Protocol
    """
    module, name = account_config.importer.rsplit(".", 1)
    try:
        cls = getattr(importlib.import_module(module), name)
    except (ImportError, AttributeError) as exc:
        raise ConfigError(
            f"Cannot import '{account_config.importer}': {exc!s}",
        ) from exc
    if not issubclass(cls, ApiImporterProtocol):
        raise ConfigError(
            f"'{account_config.importer}' is not a subclass of ApiImporterProtocol",
        )
    try:
        return cls(**account_config.model_extra)
    except (TypeError, ValueError) as exc:
        raise ConfigError(
            f"Cannot instantiate '{account_config.importer}': {exc!s}",
        ) from exc



"""Tests of the clerk module.

TODO:
    * Some tests are rather incomplete or a mess (mostly sanity only;
    multiple tests, share the same test data). Improve them.
    * Test fn append_entry_to_file.
"""
from datetime import date
from decimal import Decimal
from pathlib import Path

import pytest
import rich
from beancount.core.data import Amount, Transaction
from beancount.loader import load_file

from beanclerk.bean_helpers import create_posting, create_transaction
from beanclerk.clerk import (
    categorize,
    compute_balance,
    find_categorization_rule,
    find_last_import_date,
    import_transactions,
    transaction_exists,
)
from beanclerk.config import Config, load_config

CZK = "CZK"


@pytest.fixture()
def entries() -> list[Transaction]:
    """Return a list of Beancount entries."""
    account = "Assets:Dummy"
    postings = [create_posting(account, Amount(Decimal(1), CZK))]
    return [
        create_transaction(
            date(2023, 1, 1),
            meta={"id": "0", "ks": "0558"},
            postings=postings,
        ),
        create_transaction(date(2023, 1, 2), meta={"id": "1"}, postings=postings),
        create_transaction(date(2023, 1, 3), postings=postings),
    ]


def test_find_last_import_date(entries: list[Transaction]) -> None:
    """Test find_last_import_date."""
    account = entries[0].postings[0].account
    assert find_last_import_date([], account) is None
    assert find_last_import_date(entries, account) == date(2023, 1, 2) is not None
    assert find_last_import_date(list(entries[-1]), account) is None
    assert find_last_import_date(entries, "Assets:Nonexistent") is None


def test_transaction_exists(entries: list[Transaction]) -> None:
    """Test transaction_exists."""
    account = entries[0].postings[0].account
    assert not transaction_exists([], account, "0")
    assert transaction_exists(entries, account, "0") is not None
    assert not transaction_exists(entries, account, "-1")


@pytest.fixture()
def config(config_file: Path, ledger: Path) -> Config:
    """Return a Beanclerk Config object."""
    return load_config(config_file)


@pytest.fixture()
def _mock_prompt(monkeypatch) -> None:
    """Mock rich.Prompt.ask."""

    def mock_ask(*args, **kwargs):
        return "i"

    monkeypatch.setattr(rich.prompt.Prompt, "ask", mock_ask)


@pytest.mark.usefixtures("_mock_prompt")
def test_find_categorization_rule(config: Config, entries: list[Transaction]) -> None:
    """Test find_categorization_rule."""
    rule_1 = find_categorization_rule(entries[0], config)
    assert rule_1 is not None
    assert rule_1.account == "Expenses:Todo"
    assert rule_1.payee == "My payee"

    rule_2 = find_categorization_rule(entries[1], config)
    assert rule_2 is None


@pytest.mark.usefixtures("_mock_prompt")
def test_categorize(config: Config, entries: list[Transaction]):
    """Test categorize."""
    txn_1 = categorize(entries[0], config)
    assert txn_1.payee == "My payee"
    assert any("Expenses:Todo" in p.account for p in txn_1.postings)

    txn_2 = categorize(entries[1], config)
    assert txn_2.payee is None
    assert not any("Expenses:Todo" in p.account for p in txn_2.postings)


def test_compute_balance(entries: list[Transaction]):
    """Test compute_balance."""
    account = entries[0].postings[0].account
    assert compute_balance(entries, account, CZK) == Amount(Decimal(3), CZK)
    assert compute_balance(entries, "Assets:Nonexistent", CZK) == Amount(
        Decimal(0),
        CZK,
    )
    with pytest.raises(ValueError, match="not a valid Beancount account"):
        compute_balance(entries, "Invalid", CZK)
    with pytest.raises(ValueError, match="not a valid currency"):
        compute_balance(entries, account, "Invalid")


@pytest.mark.usefixtures("_mock_fio_banka", "_mock_prompt")
def test_import_transactions(config_file: Path, ledger: Path):
    """Test import_transactions."""
    account = "Assets:Banks:Fio:Checking"
    entries, errors, _ = load_file(ledger)
    assert not errors
    assert compute_balance(entries, account, CZK) == Amount(
        Decimal("1000.99"),
        CZK,
    )
    import_transactions(
        config_file,
        from_date=date(2023, 1, 1),
        to_date=date(2023, 1, 1),
    )

    # Don't check for errors, some entries are unbalanced due to
    # _mock_prompt fixture behavior.
    entries, _, _ = load_file(ledger)

    assert compute_balance(entries, account, CZK) == Amount(
        Decimal("2000.10"),
        CZK,
    )
    for txn_id in ("10000000000", "10000000001", "10000000002"):
        assert transaction_exists(entries, account, txn_id)



"""Clerk operations

This module provides operations consumed by the CLI.

TODO:
    * handle exceptions
    * Python docs recommend to use utf-8 encoding for reading and writing files
    * validate txns coming from importers:
        * check that txns have only 1 posting
        * check that txns have id in their metadata
    * Test fn append_entry_to_file.
"""

import copy
import re
from datetime import date
from decimal import Decimal
from pathlib import Path

from beancount.core.data import Amount, Directive, Transaction, TxnPosting
from beancount.core.realization import compute_postings_balance, postings_by_account
from beancount.loader import load_file
from beancount.parser.printer import print_entry
from rich import print as rprint
from rich.prompt import Prompt

from .bean_helpers import (
    check_account_name,
    create_posting,
    create_transaction,
    filter_entries,
)
from .config import CategorizationRule, Config, load_config, load_importer
from .exceptions import ClerkError
from .importers import ApiImporterProtocol


def find_last_import_date(entries: list[Directive], account_name: str) -> date | None:
    """Return date of the last imported transaction, or None if not found.

    This function searches for the latest transaction with `id` key in its
    metadata. Entries must be properly ordered.

    Args:
        entries (list[Directive]): a list of Beancount directives
        account_name (str): Beancount account name

    Raises:
        ValueError: if account_name is not a valid Beancount account name

    Returns:
        date | None
    """
    check_account_name(account_name)
    txn_postings = filter_entries(
        postings_by_account(entries)[account_name],
        TxnPosting,
    )
    for txn_posting in reversed(list(txn_postings)):  # latest first
        if txn_posting.txn.meta.get("id") is not None:
            return txn_posting.txn.date
    return None


def transaction_exists(
    entries: list[Directive],
    account_name: str,
    txn_id: str,
) -> bool:
    """Return True if the account has a transaction with the given ID.

    Args:
        entries (list[Directive]): a list of Beancount directives
        account_name (str): Beancount account name
        txn_id (str): transaction ID (`id` key in its metadata)

    Raises:
        ValueError: if account_name is not a valid Beancount account name

    Returns:
        bool
    """
    check_account_name(account_name)
    txn_postings = filter_entries(
        postings_by_account(entries)[account_name],
        TxnPosting,
    )
    return any(txn_posting.txn.meta.get("id") == txn_id for txn_posting in txn_postings)


def compute_balance(
    entries: list[Directive],
    account_name: str,
    currency: str,
) -> Amount:
    """Return account balance for the given account and currency.

    If the account does not exist, it returns Amount 0.

    Args:
        entries (list[Directive]): a list of Beancount directives
        account_name (str): Beancount account name
        currency (str): currency ISO code (e.g. 'USD')

    Raises:
        ValueError: if account_name is not a valid Beancount account name
        ValueError: if currency is not a valid ISO code

    Returns:
        Amount: account balance
    """
    check_account_name(account_name)
    if not re.match(r"^[A-Z]{3}$", currency):
        raise ValueError(f"'{currency}' is not a valid currency code")
    return compute_postings_balance(
        postings_by_account(entries)[account_name],
    ).get_currency_units(currency)


def find_categorization_rule(
    transaction: Transaction,
    config: Config,
) -> CategorizationRule | None:
    """Return a categorization rule matching the given transaction.

    If no rule matches the transaction, the user is prompted to choose
    an action to resolve the situation. The user may also choose not
    to categorize the transaction, None is returned then.

    Args:
        transaction (Transaction): a Beancount transaction
        config (Config): Beanclerk config

    Raises:
        ClerkError: if a categorization rule is invalid.
        ClerkError: if a dangerous pattern is used in a categorization rule.
        ClerkError: if an unexpected action is chosen by the user.

    Returns:
        CategorizationRule | None: a matching rule, or None
    """
    while True:
        if config.categorization_rules:
            for rule in config.categorization_rules:
                if len(rule.matches.metadata) == 0:
                    raise ClerkError(
                        f"Categorization rule: {rule}\n"
                        "Sanity check failed: no patterns to match",
                    )
                num_matches = 0
                for key, pattern in rule.matches.metadata.items():
                    if pattern == "":
                        raise ClerkError(
                            f"Categorization rule: {rule}\n"
                            'Dangerous pattern "" matches everything. '
                            'Use ".*" or "^$" instead.',
                        )
                    if (
                        key in transaction.meta
                        and re.search(pattern, transaction.meta[key]) is not None
                    ):
                        num_matches += 1
                if num_matches == len(rule.matches.metadata):
                    return rule

        rprint("No categorization rule matches the following transaction:")
        rprint(transaction)
        rprint("Available actions:")
        rprint("'r': reload config (you should add a new rule first)")
        rprint("'i': import as-is (transaction remains unbalanced)")
        match Prompt.ask("Enter the action", choices=["r", "i"]):
            case "r":
                # Reload only the categorization rules, changing the other
                # parts of the config may cause unexpected issues down
                # the road.
                config.categorization_rules = load_config(
                    config.config_file,
                ).categorization_rules
                continue
            case "i":
                break
            case _ as action:
                raise ClerkError(f"Unknown action: {action}")
    return None


def categorize(transaction: Transaction, config: Config) -> Transaction:
    """Return transaction categorized according to rules set in config.

    Categorization means adding any missing postings (legs) to a transaction
    to make it balanced. It may also fill in a missing payee, narration or
    transaction flag.

    The rules are applied in the order they are defined in the config file.

    The returned transaction is either a new instance (if new data have
    been added), or the original one if no matching categorization rule was
    found.

    Args:
        transaction (Transaction): a Beancount transaction
        config (Config): Beanclerk config

    Side effects:
        * `config.categorization_rules` may be modified if the user chooses
        to manually edit and reload the config file during the interactive
        categorization process.

    Returns:
        Transaction: a Beancount transaction
    """
    rule = find_categorization_rule(transaction, config)
    if rule is None:
        return transaction
    # Do categorize (Transaction is immutable, so we need to create a new one)
    units = transaction.postings[0].units
    postings = copy.deepcopy(transaction.postings)
    postings.append(
        create_posting(
            account=rule.account,
            units=Amount(-units.number, units.currency),
        ),
    )
    return create_transaction(
        _date=transaction.date,
        flag=rule.flag if rule.flag is not None else transaction.flag,
        payee=rule.payee if rule.payee is not None else transaction.payee,
        narration=rule.narration
        if rule.narration is not None
        else transaction.narration,
        meta=transaction.meta,
        postings=postings,
    )


def append_entry_to_file(entry: Directive, filepath: Path) -> None:
    """Append an entry to a file.

    Args:
        entry (Directive): a Beancount directive
        filepath (Path): a file path
    """
    with filepath.open("r") as f:
        lines = f.readlines()
        last_line = lines[-1] if lines else ""
    with filepath.open("a") as f:
        if last_line == "\n":
            pass
        elif not last_line.endswith("\n"):
            f.write(2 * "\n")
        else:
            f.write("\n")
        print_entry(entry, file=f)


def print_import_status(
    new_txns: int,
    importer_balance: Decimal,
    bean_balance: Decimal,
) -> None:
    """Print import status to stdout.

    Args:
        new_txns (int): number of imported transactions
        importer_balance (Decimal): balance reported by the importer
        bean_balance (Decimal): balance computed from the Beancount input file
    """
    diff = importer_balance - bean_balance
    if diff == 0:
        balance_status = f"OK: {importer_balance}"
    else:
        balance_status = f"NOT OK: {importer_balance} (diff: {diff})"
    rprint(f"New transactions: {new_txns}, balance {balance_status}")


def import_transactions(
    config_file: Path,
    from_date: date | None,
    to_date: date | None,
) -> None:
    """For each configured importer, import transactions and print import status.

    Args:
        config_file (Path): path to a config file
        from_date (date | None): the first date to import
        to_date (date | None): the last date to import

    Raises:
        ClerkError: raised if there are errors in the input file
        ClerkError: raised if the initial import date cannot be determined
    """
    """For each configured importer, import transactions and print import status."""
    config = load_config(config_file)
    entries, errors, _ = load_file(config.input_file)
    if errors != []:
        # TODO: format errors via beancount.parser.printer.format_errors
        raise ClerkError(f"Errors in the input file: {errors}")

    for account_config in config.accounts:
        rprint(f"Importing transactions for account: '{account_config.account}'")
        if from_date is None:
            # TODO: sort entries by date
            last_date = find_last_import_date(entries, account_config.account)
            if last_date is None:
                # TODO: catch and add a note the user should use --from-date option
                raise ClerkError("Cannot determine the initial import date.")
            from_date = last_date
        if to_date is None:
            # Beancount does not work with times, `date.today()` should be OK.
            to_date = date.today()  # noqa: DTZ011
        importer: ApiImporterProtocol = load_importer(account_config)
        txns, balance = importer.fetch_transactions(
            bean_account=account_config.account,
            from_date=from_date,
            to_date=to_date,
        )

        new_txns = 0
        for txn in txns:
            if transaction_exists(entries, account_config.account, txn.meta["id"]):
                continue
            new_txns += 1
            txn = categorize(txn, config)  # noqa: PLW2901
            append_entry_to_file(txn, config.input_file)

            # HACK: Update the list of entries without reloading the whole input
            #   file (it may be a quite slow with the Beancount v2). This way
            #   entries become unsorted and potentially unbalanced, but for
            #   a simple balance check it should be OK.
            entries.append(txn)

        print_import_status(
            new_txns,
            balance.number,
            compute_balance(entries, account_config.account, balance.currency).number,
        )
